<?php
namespace APS;

use com\parallels\aps\types\pa\account\paAccount;

require_once "aps/2/uuid.php";
require_once "aps/2/rest.php";

class Proto
{
	var $apsc;
	
	const GET    = "GET";
	const PUT    = "PUT";
	const POST   = "POST";
	const DELETE = "DELETE";
	
	// APS HTTP headers
	const APS_CONTROLLER_URI 	= "APS-Controller-URI";
	const APS_SESSION 			= "APS-Transaction-ID";
    const APS_RESOURCE_ID		= "APS-Resource-ID";
	const APS_REQUEST_ID		= "APS-Request-ID";
    const APS_TOKEN		        = "APS-Token";

    const PATH_SUBSCRIPTIONS	= "aps/subscriptions";
    const RESOURCE_LOG_EVENTS	= "aps/log";
    const APPLICATION			= "aps/2/application/";
		
	public function __construct(ControllerProxy $apsc)
	{
		$this->apsc = $apsc;
	}
	
	public static function getControllerURI()
	{
		if (key_exists("HTTP_APS_CONTROLLER_URI", $_SERVER)) {
				return $_SERVER["HTTP_APS_CONTROLLER_URI"];
		}
		if(isset($_SERVER['HTTP_APS_INSTANCE_ID'])) {
			return json_decode(file_get_contents(TypeLibrary::getConfFile()))->controllerURI;			
		}
	}
	
	public static function getSession()
	{
		if (key_exists("HTTP_APS_TRANSACTION_ID", $_SERVER))
				return $_SERVER["HTTP_APS_TRANSACTION_ID"];
		return null;
	}

	public function typesPath($id = NULL)
	{		
		$prefix = 'aps/2/types';

		if (!$id)
			return $prefix;

		return "$prefix?id=$id";
	}
	
	public static function resourcePath($id = NULL, $suffix = NULL)
	{		
		$prefix = 'aps/2/resources';

		if (!$id)
			return $prefix;

		if ($id instanceof ResourceBase)
			$id = $id->aps->id;

		if($suffix)
			return "$prefix/$id/$suffix";

		return "$prefix/$id";
	}
	
	public function getHeaders()
	{
		$h = array(
					self::APS_CONTROLLER_URI 	=> $this->apsc->getEndPoint(),
					self::APS_REQUEST_ID 		=> UUID::v4(),
					'Content-type'				=> 'application/json',
					'Accept'					=> '*/*'	
				    );
				    
		if ($this->apsc->getSession())
			$h[self::APS_SESSION] = $this->apsc->getSession();

        if ($apsToken = $this->apsc->getToken())
            $h[self::APS_TOKEN] = $apsToken;

        if ($resourceId = $this->apsc->getResourceId())
            $h[self::APS_RESOURCE_ID] = $resourceId;
        
        if ($additionalHeaders = $this->apsc->getAdditionalHeaders())
            $h = array_merge($h, $additionalHeaders);

		$headers = array();
		foreach ($h as $k => $v) 
			$headers[] = "$k: $v";
			
		return $headers;
	}

	public function sendRequest($verb, $path, $input = NULL)
	{
		$uri = $this->apsc->getEndPoint() . $path;

        $params = array(
            'http' => array(
                'method' => $verb,
                'content' => $input,
                'header' => $this->getHeaders(),
                'ignore_errors' => true,
            )
        );

        if(strncmp($uri, "https", 5) == 0)
            $params['ssl']['local_cert'] = TypeLibrary::getConfFile().".pem";

        if (defined("APS_DEVELOPMENT_MODE")) {
            Logger::get()->debug("Outgoing APS Controller API request: $verb $uri");
            Logger::get()->debug("Request:\n$input");
        }

        try {
            $ctx = stream_context_create($params);
            $fp = fopen($uri, 'rb', false, $ctx);
            if (!$fp) {
                throw new \Exception("Failed to connect to APS controller");
            }

            $httpCode = 200;
            $metaData = stream_get_meta_data($fp);
            foreach ($metaData['wrapper_data'] as $data) {
                if (preg_match('/^HTTP\/\S+ (\d+)/', $data, $matches)) {
                    $httpCode = $matches[1];
                    break;
                }
            }

            $response = stream_get_contents($fp);

            if (defined("APS_DEVELOPMENT_MODE"))
                Logger::get()->debug("Response:\n$response");

            if ($httpCode >= 300) {
                if ('{' == $response[0]) {
                    // Try to extract error if exists (JSON only)
                    $errorObj = json_decode($response);
                    if (isset($errorObj->code) && isset($errorObj->type) && isset($errorObj->message)) {
                        // Error found, throw Rest Api error
                        throw new \Rest\RestException($errorObj->code, $errorObj->type . ': ' . $errorObj->message);
                    }
                }
                throw new \Rest\RestException($httpCode, $response);
            }
        } catch (\Exception $e) {
            throw new \Exception("Request to APS controller failed: " . $e->getMessage());
        }
		return $response;
	}
}

class ControllerProxy
{

	protected $_endPoint;
    protected $_session;
    protected $_token;
    protected $_resourceId;

    /** @var array */
    protected $_headers;
	protected $_io;
	
	public function __construct($endPoint, $session = null, $token = null)
	{
		if (!$endPoint)
				throw new \Exception("Can't determine controller end-point");	
		
		if (substr($endPoint, -1) != "/")
			$endPoint .= "/";

		$this->_endPoint = $endPoint;
		$this->_session = $session;
		$this->_io = new Proto($this);
        $this->_token = $token;
		
		// TODO: from incoming request fill-in: session, endPoint
	}
	
	public function __clone()
	{
		$this->_io = new Proto($this);
	}
	
	public static function listInstances() {
		$instances = array();
		foreach(glob('config'.DIRECTORY_SEPARATOR.'*') as $instance) {
			if(!strpos($instance, '.')) {
				$parts = explode(DIRECTORY_SEPARATOR, $instance);
				$instances[] = end($parts);
			}
		}
		return $instances;
	}
	
	public function resetSession()
	{
		$this->_session = null;
	}
	
	public function getSession()
	{
		return $this->_session;
	}
	
	public function getEndPoint()
	{
		return $this->_endPoint;
	}

    public function getToken()
    {
        return $this->_token;
	}

	
	public function setResourceId($resourceId)
	{
		$this->_resourceId = $resourceId;
	}
	
	public function getResourceId()
	{
		return $this->_resourceId;
	}

    public function setAdditionalHeaders(array $headers)
    {
        $this->_headers = $headers;
    }

    public function getAdditionalHeaders()
    {
        return $this->_headers;
    }

    public function getType($type)
    {
        $path = $this->getIo()->typesPath($type);
        $representations = json_decode($this->getIo()->sendRequest('GET', $path));
        if (empty($representations)) {
            throw new SchemaException("Failed to find class '$type' by type id");
        }
        return $representations[0];
    }

	public function getResources($rqlFilter = null, $path = null)
	{
		if(!$path)
			$path = $this->_io->resourcePath();
		$out = $this->_io->sendRequest(Proto::GET, $path."?$rqlFilter");
		$resources = array();
		if ($out[0] == '<')
		{
			$dom = new \DOMDocument();
			$dom->loadXML($out);
			$root = $dom->documentElement;
			if ($root->tagName != "resources")
				throw new \Exception("Incorrect type");
			foreach($root->childNodes as $node)
			{
				if ($node->nodeType != \XML_ELEMENT_NODE)
					continue;
					
				try
				{
					$resources[] = TypeLibrary::newResourceXML($node);
				} 
				catch(\Exception $e) {}; // TODO: do not panic due to unknown resources
			}	
		} 
		else 
		{
			foreach(json_decode($out) as $node)
				try
				{
					$resources[] = TypeLibrary::newResourceJSON($node);
				} 
				catch(\Exception $e) {}; // TODO: do not panic due to unknown resources
		}
		return $resources;		
	}
	
	public function log(ResourceBase $resource, $severity, $message, array $parameters = NULL, $time = NULL) {
		return $this->logById($resource->aps->id, $severity, $message, $parameters, $time);
	}
		
	public function logById($resourceId, $severity, $message, array $parameters = NULL, $time = NULL) {
		static $levels  = array('critical', 'error', 'warning', 'notice', 'info', 'debug');
		if(!in_array($severity, $levels)) {
			Logger::get()->info("Fail to process log. Invalid severity level");
			return NULL;
		}
		$logObject = new \stdClass();
		$logObject->message = $message;
		$logObject->severity = $severity;		
		if($time)
			$logObject->time = time;
		$logObject->parameters = $parameters;
		$response = $this->_io->sendRequest(Proto::POST, $this->_io->resourcePath($resourceId, PROTO::RESOURCE_LOG_EVENTS), json_encode($logObject));
		return json_decode($response);
	}

	public function getLogs(ResourceBase $resource) {
		return $this->getLogsById($resource->aps->id);
	}
		
	public function getLogsById($resourceId) {
		return $this->_io->sendRequest(Proto::GET, $this->_io->resourcePath($resourceId, PROTO::RESOURCE_LOG_EVENTS));
	}
	
	public function deleteLog(ResourceBase $resource) {
		return $this->deleteLogById($resource->aps->id);
	}
	
	public function deleteLogById($resourceId) {
		return $this->_io->sendRequest(Proto::DELETE, $this->_io->resourcePath($resourceId, PROTO::RESOURCE_LOG_EVENTS));		
	}
	
    public function getIo()
    {
        return $this->_io;
    }
	
	public function getResource($id)
	{
		if (!$id)
			throw new \Exception("No resourceId provided to getResources");

		$out = $this->_io->sendRequest(Proto::GET, $this->_io->resourcePath($id));
		$resource = TypeLibrary::parseResource($out);  
		return $resource;
	}		

    public function updateResource(ResourceBase $newResource)
	{
		$id = $newResource->aps->id;
        $type = $newResource->aps->type;
        $service = TypeLibrary::getSchemaByType($type)->service;
        return $this->_io->sendRequest(Proto::PUT, Proto::APPLICATION."$service/$id", $newResource->toString());
	}
	
    public function registerResource(ResourceBase $resource)	{
        $service = TypeLibrary::getSchemaByType($resource->aps->type)->service;
		$out = $this->_io->sendRequest(Proto::POST, Proto::APPLICATION.$service, $resource->toString());
		$ret = TypeLibrary::parseResource($out);
		return $ret;
	}	

    public function unregisterResource(ResourceBase $resource)
	{
        $service = TypeLibrary::getSchemaByType($resource->aps->type)->service;
        $this->_io->sendRequest(Proto::DELETE, Proto::APPLICATION."$service/".$resource->aps->id);
	}
	
	public function subscribe(ResourceBase $base, EventSubscription $sub)
	{
		if ($sub->id)
			$out = $this->_io->sendRequest(
				Proto::PUT, 
				$this->_io->resourcePath($base->aps->id, Proto::PATH_SUBSCRIPTIONS . $sub->id), 
				$sub->toJSON());
		else
			$out = $this->_io->sendRequest(
				Proto::POST, 
				$this->_io->resourcePath($base->aps->id, Proto::PATH_SUBSCRIPTIONS), 
				$sub->toJSON());
		
		$ret = new EventSubscription();
		$ret->fromJSON($out);
		
		return $ret;
	}
	
	public function unsubscribe(ResourceBase $base, EventSubscription $sub)
	{
		$out = $this->_io->sendRequest(
				Proto::DELETE, 
				$this->_io->resourcePath($base->aps->id, Proto::PATH_SUBSCRIPTIONS ."/" . $sub->id));
	}

	public function subscriptions(ResourceBase $base)
	{
		$out = $this->_io->sendRequest(
				Proto::GET, 
				$this->_io->resourcePath($base->aps->id, Proto::PATH_SUBSCRIPTIONS));
		
		$ret = array();
		foreach (json_decode($out) as $i)
		{
			$sub = new EventSubscription();
			$ret[] = $sub->fromJSON($i);
		}
		
		return $ret;
	}
}
